import copy
import inspect
from functools import partial
from collections import OrderedDict

import six

from ..exceptions import SkipField
from .options import Options


class ClassTypeMeta(type):
    options_class = Options

    def __new__(mcs, name, bases, attrs):
        super_new = super(ClassTypeMeta, mcs).__new__

        module = attrs.pop('__module__', None)
        doc = attrs.pop('__doc__', None)
        new_class = super_new(mcs, name, bases, {
            '__module__': module,
            '__doc__': doc
        })
        attr_meta = attrs.pop('Meta', None)
        if not attr_meta:
            meta = getattr(new_class, 'Meta', None)
        else:
            meta = attr_meta

        new_class.add_to_class('_meta', new_class.get_options(meta))

        return mcs.construct(new_class, bases, attrs)

    def get_options(cls, meta):
        return cls.options_class(meta)

    def add_to_class(cls, name, value):
        # We should call the contribute_to_class method only if it's bound
        if not inspect.isclass(value) and hasattr(
                value, 'contribute_to_class'):
            value.contribute_to_class(cls, name)
        else:
            setattr(cls, name, value)

    def construct(cls, bases, attrs):
        # Add all attributes to the class.
        for obj_name, obj in attrs.items():
            cls.add_to_class(obj_name, obj)

        if not cls._meta.abstract:
            from ..types import List, NonNull
            setattr(cls, 'NonNull', partial(NonNull, cls))
            setattr(cls, 'List', partial(List, cls))

        return cls


class ClassType(six.with_metaclass(ClassTypeMeta)):

    class Meta:
        abstract = True

    @classmethod
    def internal_type(cls, schema):
        raise NotImplementedError("Function internal_type not implemented in type {}".format(cls))


class FieldsOptions(Options):

    def __init__(self, *args, **kwargs):
        super(FieldsOptions, self).__init__(*args, **kwargs)
        self.local_fields = []

    def add_field(self, field):
        self.local_fields.append(field)

    @property
    def fields(self):
        return sorted(self.local_fields)

    @property
    def fields_map(self):
        return OrderedDict([(f.attname, f) for f in self.fields])


class FieldsClassTypeMeta(ClassTypeMeta):
    options_class = FieldsOptions

    def extend_fields(cls, bases):
        new_fields = cls._meta.local_fields
        field_names = {f.name: f for f in new_fields}

        for base in bases:
            if not isinstance(base, FieldsClassTypeMeta):
                continue

            parent_fields = base._meta.local_fields
            for field in parent_fields:
                if field.name in field_names and field.type.__class__ != field_names[
                        field.name].type.__class__:
                    raise Exception(
                        'Local field %r in class %r (%r) clashes '
                        'with field with similar name from '
                        'Interface %s (%r)' % (
                            field.name,
                            cls.__name__,
                            field.__class__,
                            base.__name__,
                            field_names[field.name].__class__)
                    )
                new_field = copy.copy(field)
                cls.add_to_class(field.attname, new_field)

    def construct(cls, bases, attrs):
        cls = super(FieldsClassTypeMeta, cls).construct(bases, attrs)
        cls.extend_fields(bases)
        return cls


class FieldsClassType(six.with_metaclass(FieldsClassTypeMeta, ClassType)):

    class Meta:
        abstract = True

    @classmethod
    def fields_internal_types(cls, schema):
        fields = []
        for field in cls._meta.fields:
            try:
                fields.append((field.name, schema.T(field)))
            except SkipField:
                continue

        return OrderedDict(fields)



from functools import partial

from graphql.core.type import GraphQLInputObjectType

from .base import FieldsClassType


class InputObjectType(FieldsClassType):

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        raise Exception("An InputObjectType cannot be initialized")

    @classmethod
    def internal_type(cls, schema):
        if cls._meta.abstract:
            raise Exception("Abstract InputObjectTypes don't have a specific type.")

        return GraphQLInputObjectType(
            cls._meta.type_name,
            description=cls._meta.description,
            fields=partial(cls.fields_internal_types, schema),
        )




from graphql.core.type import GraphQLInputObjectType

from graphene.core.schema import Schema
from graphene.core.types import String

from ..inputobjecttype import InputObjectType


def test_inputobjecttype():
    class InputCharacter(InputObjectType):
        '''InputCharacter description'''
        name = String()

    schema = Schema()

    object_type = schema.T(InputCharacter)
    assert isinstance(object_type, GraphQLInputObjectType)
    assert InputCharacter._meta.type_name == 'InputCharacter'
    assert object_type.description == 'InputCharacter description'
    assert list(object_type.get_fields().keys()) == ['name']



from graphql.core.type import GraphQLScalarType

from ..types.base import MountedType
from .base import ClassType


class Scalar(ClassType, MountedType):

    @classmethod
    def internal_type(cls, schema):
        serialize = getattr(cls, 'serialize')
        parse_literal = getattr(cls, 'parse_literal')
        parse_value = getattr(cls, 'parse_value')

        return GraphQLScalarType(
            name=cls._meta.type_name,
            description=cls._meta.description,
            serialize=serialize,
            parse_value=parse_value,
            parse_literal=parse_literal
        )



class Options(object):

    def __init__(self, meta=None, **defaults):
        self.meta = meta
        self.abstract = False
        for name, value in defaults.items():
            setattr(self, name, value)
        self.valid_attrs = list(defaults.keys()) + ['type_name', 'description', 'abstract']

    def contribute_to_class(self, cls, name):
        cls._meta = self
        self.parent = cls
        # First, construct the default values for these options.
        self.object_name = cls.__name__
        self.type_name = self.object_name

        self.description = cls.__doc__
        # Store the original user-defined values for each option,
        # for use when serializing the model definition
        self.original_attrs = {}

        # Next, apply any overridden values from 'class Meta'.
        if self.meta:
            meta_attrs = self.meta.__dict__.copy()
            for name in self.meta.__dict__:
                # Ignore any private attributes that Django doesn't care about.
                # NOTE: We can't modify a dictionary's contents while looping
                # over it, so we loop over the *original* dictionary instead.
                if name.startswith('_'):
                    del meta_attrs[name]
            for attr_name in self.valid_attrs:
                if attr_name in meta_attrs:
                    setattr(self, attr_name, meta_attrs.pop(attr_name))
                    self.original_attrs[attr_name] = getattr(self, attr_name)
                elif hasattr(self.meta, attr_name):
                    setattr(self, attr_name, getattr(self.meta, attr_name))
                    self.original_attrs[attr_name] = getattr(self, attr_name)

            del self.valid_attrs

            # Any leftover attributes must be invalid.
            if meta_attrs != {}:
                raise TypeError(
                    "'class Meta' got invalid attribute(s): %s" %
                    ','.join(
                        meta_attrs.keys()))

        del self.meta



from graphql.core.type import GraphQLInterfaceType, GraphQLObjectType
from py.test import raises

from graphene.core.schema import Schema
from graphene.core.types import String

from ..interface import Interface
from ..objecttype import ObjectType


def test_interface():
    class Character(Interface):
        '''Character description'''
        name = String()

    schema = Schema()

    object_type = schema.T(Character)
    assert issubclass(Character, Interface)
    assert isinstance(object_type, GraphQLInterfaceType)
    assert Character._meta.interface
    assert Character._meta.type_name == 'Character'
    assert object_type.description == 'Character description'
    assert list(object_type.get_fields().keys()) == ['name']


def test_interface_cannot_initialize():
    class Character(Interface):
        pass

    with raises(Exception) as excinfo:
        Character()
    assert 'An interface cannot be initialized' == str(excinfo.value)


def test_interface_inheritance_abstract():
    class Character(Interface):
        pass

    class ShouldBeInterface(Character):

        class Meta:
            abstract = True

    class ShouldBeObjectType(ShouldBeInterface):
        pass

    assert ShouldBeInterface._meta.interface
    assert not ShouldBeObjectType._meta.interface
    assert issubclass(ShouldBeObjectType, ObjectType)


def test_interface_inheritance():
    class Character(Interface):
        pass

    class GeneralInterface(Interface):
        pass

    class ShouldBeObjectType(GeneralInterface, Character):
        pass

    schema = Schema()

    assert Character._meta.interface
    assert not ShouldBeObjectType._meta.interface
    assert issubclass(ShouldBeObjectType, ObjectType)
    assert Character in ShouldBeObjectType._meta.interfaces
    assert GeneralInterface in ShouldBeObjectType._meta.interfaces
    assert isinstance(schema.T(Character), GraphQLInterfaceType)
    assert isinstance(schema.T(ShouldBeObjectType), GraphQLObjectType)


def test_interface_inheritance_non_objects():
    class CommonClass(object):
        common_attr = True

    class Character(CommonClass, Interface):
        pass

    class ShouldBeObjectType(Character):
        pass

    assert Character._meta.interface
    assert Character.common_attr
    assert ShouldBeObjectType.common_attr



import inspect
import warnings
from collections import Iterable
from functools import wraps

import six

from graphql_relay.connection.arrayconnection import connection_from_list
from graphql_relay.node.node import to_global_id

from ..core.classtypes import InputObjectType, Interface, Mutation, ObjectType
from ..core.classtypes.interface import InterfaceMeta
from ..core.classtypes.mutation import MutationMeta
from ..core.types import Boolean, Field, List, String
from ..core.types.argument import ArgumentsGroup
from ..core.types.definitions import NonNull
from ..utils import memoize
from .fields import GlobalIDField


class PageInfo(ObjectType):
    has_next_page = Boolean(
        required=True,
        description='When paginating forwards, are there more items?')
    has_previous_page = Boolean(
        required=True,
        description='When paginating backwards, are there more items?')
    start_cursor = String(
        description='When paginating backwards, the cursor to continue.')
    end_cursor = String(
        description='When paginating forwards, the cursor to continue.')


class Edge(ObjectType):
    '''An edge in a connection.'''
    cursor = String(
        required=True, description='A cursor for use in pagination')

    @classmethod
    @memoize
    def for_node(cls, node):
        from graphene.relay.utils import is_node
        assert is_node(node), 'ObjectTypes in a edge have to be Nodes'
        node_field = Field(node, description='The item at the end of the edge')
        return type(
            '%s%s' % (node._meta.type_name, cls._meta.type_name),
            (cls,),
            {'node_type': node, 'node': node_field})


class Connection(ObjectType):
    '''A connection to a list of items.'''
    class Meta:
        type_name = 'DefaultConnection'

    page_info = Field(PageInfo, required=True,
                      description='The Information to aid in pagination')

    _connection_data = None

    @classmethod
    @memoize
    def for_node(cls, node, edge_type=None):
        from graphene.relay.utils import is_node
        edge_type = edge_type or Edge.for_node(node)
        assert is_node(node), 'ObjectTypes in a connection have to be Nodes'
        edges = List(edge_type, description='Information to aid in pagination.')
        return type(
            '%s%s' % (node._meta.type_name, cls._meta.type_name),
            (cls,),
            {'edge_type': edge_type, 'edges': edges})

    @classmethod
    def from_list(cls, iterable, args, info):
        assert isinstance(
            iterable, Iterable), 'Resolved value from the connection field have to be iterable'
        connection = connection_from_list(
            iterable, args, connection_type=cls,
            edge_type=cls.edge_type, pageinfo_type=PageInfo)
        connection.set_connection_data(iterable)
        return connection

    def set_connection_data(self, data):
        self._connection_data = data

    def get_connection_data(self):
        return self._connection_data


class NodeMeta(InterfaceMeta):

    def construct_get_node(cls):
        get_node = getattr(cls, 'get_node', None)
        assert get_node, 'get_node classmethod not found in %s Node' % cls
        assert callable(get_node), 'get_node have to be callable'
        args = 3
        if isinstance(get_node, staticmethod):
            args -= 1

        get_node_num_args = len(inspect.getargspec(get_node).args)
        if get_node_num_args < args:
            warnings.warn("get_node will receive also the info arg"
                          " in future versions of graphene".format(cls.__name__),
                          FutureWarning)

            @staticmethod
            @wraps(get_node)
            def wrapped_node(*node_args):
                if len(node_args) < args:
                    node_args += (None, )
                return get_node(*node_args[:-1])

            setattr(cls, 'get_node', wrapped_node)

    def construct(cls, *args, **kwargs):
        cls = super(NodeMeta, cls).construct(*args, **kwargs)
        if not cls._meta.abstract:
            cls.construct_get_node()
        return cls


class Node(six.with_metaclass(NodeMeta, Interface)):
    '''An object with an ID'''
    id = GlobalIDField()

    class Meta:
        abstract = True

    def to_global_id(self):
        type_name = self._meta.type_name
        return to_global_id(type_name, self.id)

    connection_type = Connection
    edge_type = Edge

    @classmethod
    def get_connection_type(cls):
        return cls.connection_type

    @classmethod
    def get_edge_type(cls):
        return cls.edge_type


class MutationInputType(InputObjectType):
    client_mutation_id = String(required=True)


class RelayMutationMeta(MutationMeta):

    def construct(cls, *args, **kwargs):
        cls = super(RelayMutationMeta, cls).construct(*args, **kwargs)
        if not cls._meta.abstract:
            assert hasattr(
                cls, 'mutate_and_get_payload'), 'You have to implement mutate_and_get_payload'
        return cls

    def construct_arguments(cls, items):
        new_input_type = type('{}Input'.format(
            cls._meta.type_name), (MutationInputType, ), items)
        cls.add_to_class('input_type', new_input_type)
        return ArgumentsGroup(input=NonNull(new_input_type))


class ClientIDMutation(six.with_metaclass(RelayMutationMeta, Mutation)):
    client_mutation_id = String(required=True)

    class Meta:
        abstract = True

    @classmethod
    def mutate(cls, instance, args, info):
        input = args.get('input')
        payload = cls.mutate_and_get_payload(input, info)
        client_mutation_id = input.get('client_mutation_id')
        setattr(payload, 'client_mutation_id', client_mutation_id)
        return payload




from ..data import setup
from ..schema import Schema

setup()


def test_hero_name_query():
    query = '''
        query HeroNameQuery {
          hero {
            name
          }
        }
    '''
    expected = {
        'hero': {
            'name': 'R2-D2'
        }
    }
    result = Schema.execute(query)
    assert not result.errors
    assert result.data == expected


def test_hero_name_and_friends_query():
    query = '''
        query HeroNameAndFriendsQuery {
          hero {
            id
            name
            friends {
              name
            }
          }
        }
    '''
    expected = {
        'hero': {
            'id': '2001',
            'name': 'R2-D2',
            'friends': [
                {'name': 'Luke Skywalker'},
                {'name': 'Han Solo'},
                {'name': 'Leia Organa'},
            ]
        }
    }
    result = Schema.execute(query)
    assert not result.errors
    assert result.data == expected


def test_nested_query():
    query = '''
        query NestedQuery {
          hero {
            name
            friends {
              name
              appearsIn
              friends {
                name
              }
            }
          }
        }
    '''
    expected = {
        'hero': {
            'name': 'R2-D2',
            'friends': [
                {
                    'name': 'Luke Skywalker',
                    'appearsIn': ['NEWHOPE', 'EMPIRE', 'JEDI'],
                    'friends': [
                        {
                            'name': 'Han Solo',
                        },
                        {
                            'name': 'Leia Organa',
                        },
                        {
                            'name': 'C-3PO',
                        },
                        {
                            'name': 'R2-D2',
                        },
                    ]
                },
                {
                    'name': 'Han Solo',
                    'appearsIn': ['NEWHOPE', 'EMPIRE', 'JEDI'],
                    'friends': [
                        {
                            'name': 'Luke Skywalker',
                        },
                        {
                            'name': 'Leia Organa',
                        },
                        {
                            'name': 'R2-D2',
                        },
                    ]
                },
                {
                    'name': 'Leia Organa',
                    'appearsIn': ['NEWHOPE', 'EMPIRE', 'JEDI'],
                    'friends': [
                        {
                            'name': 'Luke Skywalker',
                        },
                        {
                            'name': 'Han Solo',
                        },
                        {
                            'name': 'C-3PO',
                        },
                        {
                            'name': 'R2-D2',
                        },
                    ]
                },
            ]
        }
    }
    result = Schema.execute(query)
    assert not result.errors
    assert result.data == expected


def test_fetch_luke_query():
    query = '''
        query FetchLukeQuery {
          human(id: "1000") {
            name
          }
        }
    '''
    expected = {
        'human': {
            'name': 'Luke Skywalker',
        }
    }
    result = Schema.execute(query)
    assert not result.errors
    assert result.data == expected


def test_fetch_some_id_query():
    query = '''
        query FetchSomeIDQuery($someId: String!) {
          human(id: $someId) {
            name
          }
        }
    '''
    params = {
        'someId': '1000',
    }
    expected = {
        'human': {
            'name': 'Luke Skywalker',
        }
    }
    result = Schema.execute(query, None, params)
    assert not result.errors
    assert result.data == expected


def test_fetch_some_id_query2():
    query = '''
        query FetchSomeIDQuery($someId: String!) {
          human(id: $someId) {
            name
          }
        }
    '''
    params = {
        'someId': '1002',
    }
    expected = {
        'human': {
            'name': 'Han Solo',
        }
    }
    result = Schema.execute(query, None, params)
    assert not result.errors
    assert result.data == expected


def test_invalid_id_query():
    query = '''
        query humanQuery($id: String!) {
          human(id: $id) {
            name
          }
        }
    '''
    params = {
        'id': 'not a valid id',
    }
    expected = {
        'human': None
    }
    result = Schema.execute(query, None, params)
    assert not result.errors
    assert result.data == expected


def test_fetch_luke_aliased():
    query = '''
        query FetchLukeAliased {
          luke: human(id: "1000") {
            name
          }
        }
    '''
    expected = {
        'luke': {
            'name': 'Luke Skywalker',
        }
    }
    result = Schema.execute(query)
    assert not result.errors
    assert result.data == expected


def test_fetch_luke_and_leia_aliased():
    query = '''
        query FetchLukeAndLeiaAliased {
          luke: human(id: "1000") {
            name
          }
          leia: human(id: "1003") {
            name
          }
        }
    '''
    expected = {
        'luke': {
            'name': 'Luke Skywalker',
        },
        'leia': {
            'name': 'Leia Organa',
        }
    }
    result = Schema.execute(query)
    assert not result.errors
    assert result.data == expected


def test_duplicate_fields():
    query = '''
        query DuplicateFields {
          luke: human(id: "1000") {
            name
            homePlanet
          }
          leia: human(id: "1003") {
            name
            homePlanet
          }
        }
    '''
    expected = {
        'luke': {
            'name': 'Luke Skywalker',
            'homePlanet': 'Tatooine',
        },
        'leia': {
            'name': 'Leia Organa',
            'homePlanet': 'Alderaan',
        }
    }
    result = Schema.execute(query)
    assert not result.errors
    assert result.data == expected


def test_use_fragment():
    query = '''
        query UseFragment {
          luke: human(id: "1000") {
            ...HumanFragment
          }
          leia: human(id: "1003") {
            ...HumanFragment
          }
        }
        fragment HumanFragment on Human {
          name
          homePlanet
        }
    '''
    expected = {
        'luke': {
            'name': 'Luke Skywalker',
            'homePlanet': 'Tatooine',
        },
        'leia': {
            'name': 'Leia Organa',
            'homePlanet': 'Alderaan',
        }
    }
    result = Schema.execute(query)
    assert not result.errors
    assert result.data == expected


def test_check_type_of_r2():
    query = '''
        query CheckTypeOfR2 {
          hero {
            __typename
            name
          }
        }
    '''
    expected = {
        'hero': {
            '__typename': 'Droid',
            'name': 'R2-D2',
        }
    }
    result = Schema.execute(query)
    assert not result.errors
    assert result.data == expected


def test_check_type_of_luke():
    query = '''
        query CheckTypeOfLuke {
          hero(episode: EMPIRE) {
            __typename
            name
          }
        }
    '''
    expected = {
        'hero': {
            '__typename': 'Human',
            'name': 'Luke Skywalker',
        }
    }
    result = Schema.execute(query)
    assert not result.errors
    assert result.data == expected



from graphene.contrib.django.types import (
    DjangoConnection,
    DjangoObjectType,
    DjangoNode
)
from graphene.contrib.django.fields import (
    DjangoConnectionField,
    DjangoModelField
)

__all__ = ['DjangoObjectType', 'DjangoNode', 'DjangoConnection',
           'DjangoConnectionField', 'DjangoModelField']



import inspect
from collections import OrderedDict

from graphql.core.execution.executor import Executor
from graphql.core.execution.middlewares.sync import \
    SynchronousExecutionMiddleware
from graphql.core.type import GraphQLSchema as _GraphQLSchema
from graphql.core.utils.introspection_query import introspection_query
from graphql.core.utils.schema_printer import print_schema

from graphene import signals

from .classtypes.base import ClassType
from .types.base import BaseType


class GraphQLSchema(_GraphQLSchema):

    def __init__(self, schema, *args, **kwargs):
        self.graphene_schema = schema
        super(GraphQLSchema, self).__init__(*args, **kwargs)


class Schema(object):
    _executor = None

    def __init__(self, query=None, mutation=None, subscription=None,
                 name='Schema', executor=None):
        self._types_names = {}
        self._types = {}
        self.mutation = mutation
        self.query = query
        self.subscription = subscription
        self.name = name
        self.executor = executor
        signals.init_schema.send(self)

    def __repr__(self):
        return '<Schema: %s (%s)>' % (str(self.name), hash(self))

    def T(self, object_type):
        if not object_type:
            return
        if inspect.isclass(object_type) and issubclass(
                object_type, (BaseType, ClassType)) or isinstance(
                object_type, BaseType):
            if object_type not in self._types:
                internal_type = object_type.internal_type(self)
                self._types[object_type] = internal_type
                is_objecttype = inspect.isclass(
                    object_type) and issubclass(object_type, ClassType)
                if is_objecttype:
                    self.register(object_type)
            return self._types[object_type]
        else:
            return object_type

    @property
    def executor(self):
        if not self._executor:
            self._executor = Executor(
                [SynchronousExecutionMiddleware()], map_type=OrderedDict)
        return self._executor

    @executor.setter
    def executor(self, value):
        self._executor = value

    @property
    def schema(self):
        if not self.query:
            raise Exception('You have to define a base query type')
        return GraphQLSchema(
            self,
            query=self.T(self.query),
            mutation=self.T(self.mutation),
            subscription=self.T(self.subscription))

    def register(self, object_type):
        type_name = object_type._meta.type_name
        registered_object_type = self._types_names.get(type_name, None)
        if registered_object_type:
            assert registered_object_type == object_type, 'Type {} already registered with other object type'.format(
                type_name)
        self._types_names[object_type._meta.type_name] = object_type
        return object_type

    def objecttype(self, type):
        name = getattr(type, 'name', None)
        if name:
            objecttype = self._types_names.get(name, None)
            if objecttype and inspect.isclass(
                    objecttype) and issubclass(objecttype, ClassType):
                return objecttype

    def __str__(self):
        return print_schema(self.schema)

    def setup(self):
        assert self.query, 'The base query type is not set'
        self.T(self.query)

    def get_type(self, type_name):
        self.setup()
        if type_name not in self._types_names:
            raise KeyError('Type %r not found in %r' % (type_name, self))
        return self._types_names[type_name]

    @property
    def types(self):
        return self._types_names

    def execute(self, request='', root=None, vars=None,
                operation_name=None, **kwargs):
        root = root or object()
        return self.executor.execute(
            self.schema,
            request,
            root=root,
            args=vars,
            operation_name=operation_name,
            **kwargs
        )

    def introspect(self):
        return self.execute(introspection_query).data



from .inputobjecttype import InputObjectType
from .interface import Interface
from .mutation import Mutation
from .objecttype import ObjectType
from .options import Options
from .scalar import Scalar
from .uniontype import UnionType

__all__ = [
    'InputObjectType',
    'Interface',
    'Mutation',
    'ObjectType',
    'Options',
    'Scalar',
    'UnionType']



import graphene
from graphene import relay, resolve_only_args
from graphene.contrib.django import DjangoNode, DjangoObjectType

from .data import (create_ship, get_empire, get_faction, get_rebels, get_ship,
                   get_ships)
from .models import Character as CharacterModel
from .models import Faction as FactionModel
from .models import Ship as ShipModel

schema = graphene.Schema(name='Starwars Django Relay Schema')


class Ship(DjangoNode):

    class Meta:
        model = ShipModel

    @classmethod
    def get_node(cls, id, info):
        return Ship(get_ship(id))


class Character(DjangoObjectType):

    class Meta:
        model = CharacterModel


class Faction(DjangoNode):

    class Meta:
        model = FactionModel

    @classmethod
    def get_node(cls, id, info):
        return Faction(get_faction(id))


class IntroduceShip(relay.ClientIDMutation):

    class Input:
        ship_name = graphene.String(required=True)
        faction_id = graphene.String(required=True)

    ship = graphene.Field(Ship)
    faction = graphene.Field(Faction)

    @classmethod
    def mutate_and_get_payload(cls, input, info):
        ship_name = input.get('ship_name')
        faction_id = input.get('faction_id')
        ship = create_ship(ship_name, faction_id)
        faction = get_faction(faction_id)
        return IntroduceShip(ship=Ship(ship), faction=Faction(faction))


class Query(graphene.ObjectType):
    rebels = graphene.Field(Faction)
    empire = graphene.Field(Faction)
    node = relay.NodeField()
    ships = relay.ConnectionField(Ship, description='All the ships.')

    @resolve_only_args
    def resolve_ships(self):
        return get_ships()

    @resolve_only_args
    def resolve_rebels(self):
        return get_rebels()

    @resolve_only_args
    def resolve_empire(self):
        return get_empire()


class Mutation(graphene.ObjectType):
    introduce_ship = graphene.Field(IntroduceShip)


# We register the Character Model because if not would be
# inaccessible for the schema
schema.register(Character)

schema.query = Query
schema.mutation = Mutation



from ...core.classtypes.objecttype import ObjectTypeOptions
from ...relay.types import Node
from ...relay.utils import is_node

VALID_ATTRS = ('model', 'only_fields', 'exclude_fields')


class DjangoOptions(ObjectTypeOptions):

    def __init__(self, *args, **kwargs):
        super(DjangoOptions, self).__init__(*args, **kwargs)
        self.model = None
        self.valid_attrs += VALID_ATTRS
        self.only_fields = None
        self.exclude_fields = []

    def contribute_to_class(self, cls, name):
        super(DjangoOptions, self).contribute_to_class(cls, name)
        if is_node(cls):
            self.exclude_fields = list(self.exclude_fields) + ['id']
            self.interfaces.append(Node)



from django.db import models
from singledispatch import singledispatch

from ...core.types.scalars import ID, Boolean, Float, Int, String
from .fields import ConnectionOrListField, DjangoModelField

try:
    UUIDField = models.UUIDField
except AttributeError:
    # Improved compatibility for Django 1.6
    class UUIDField(object):
        pass


@singledispatch
def convert_django_field(field):
    raise Exception(
        "Don't know how to convert the Django field %s (%s)" %
        (field, field.__class__))


@convert_django_field.register(models.DateField)
@convert_django_field.register(models.CharField)
@convert_django_field.register(models.TextField)
@convert_django_field.register(models.EmailField)
@convert_django_field.register(models.SlugField)
@convert_django_field.register(models.URLField)
@convert_django_field.register(UUIDField)
def convert_field_to_string(field):
    return String(description=field.help_text)


@convert_django_field.register(models.AutoField)
def convert_field_to_id(field):
    return ID(description=field.help_text)


@convert_django_field.register(models.PositiveIntegerField)
@convert_django_field.register(models.PositiveSmallIntegerField)
@convert_django_field.register(models.SmallIntegerField)
@convert_django_field.register(models.BigIntegerField)
@convert_django_field.register(models.IntegerField)
def convert_field_to_int(field):
    return Int(description=field.help_text)


@convert_django_field.register(models.BooleanField)
def convert_field_to_boolean(field):
    return Boolean(description=field.help_text, required=True)


@convert_django_field.register(models.NullBooleanField)
def convert_field_to_nullboolean(field):
    return Boolean(description=field.help_text)


@convert_django_field.register(models.DecimalField)
@convert_django_field.register(models.FloatField)
def convert_field_to_float(field):
    return Float(description=field.help_text)


@convert_django_field.register(models.ManyToManyField)
@convert_django_field.register(models.ManyToOneRel)
def convert_field_to_list_or_connection(field):
    model_field = DjangoModelField(field.related_model)
    return ConnectionOrListField(model_field)


@convert_django_field.register(models.OneToOneField)
@convert_django_field.register(models.ForeignKey)
def convert_field_to_djangomodel(field):
    return DjangoModelField(field.related_model, description=field.help_text)



import graphene


class GeoInput(graphene.InputObjectType):
    lat = graphene.Float(required=True)
    lng = graphene.Float(required=True)


class Address(graphene.ObjectType):
    latlng = graphene.String()


class Query(graphene.ObjectType):
    address = graphene.Field(Address, geo=graphene.Argument(GeoInput))

    def resolve_address(self, args, info):
        geo = args.get('geo')
        return Address(latlng="({},{})".format(geo.get('lat'), geo.get('lng')))


schema = graphene.Schema(query=Query)
query = '''
    query something{
      address(geo: {lat:32.2, lng:12}) {
        latlng
      }
    }
'''

result = schema.execute(query)
print(result.data['address']['latlng'])



from ...schema import Schema
from ...types import Field, List, NonNull, String
from ..base import ClassType, FieldsClassType


def test_classtype_basic():
    class Character(ClassType):
        '''Character description'''
    assert Character._meta.type_name == 'Character'
    assert Character._meta.description == 'Character description'


def test_classtype_advanced():
    class Character(ClassType):

        class Meta:
            type_name = 'OtherCharacter'
            description = 'OtherCharacter description'
    assert Character._meta.type_name == 'OtherCharacter'
    assert Character._meta.description == 'OtherCharacter description'


def test_classtype_definition_list():
    class Character(ClassType):
        '''Character description'''
    assert isinstance(Character.List(), List)
    assert Character.List().of_type == Character


def test_classtype_definition_nonnull():
    class Character(ClassType):
        '''Character description'''
    assert isinstance(Character.NonNull(), NonNull)
    assert Character.NonNull().of_type == Character


def test_fieldsclasstype_definition_order():
    class Character(ClassType):
        '''Character description'''

    class Query(FieldsClassType):
        name = String()
        char = Character.NonNull()

    assert list(Query._meta.fields_map.keys()) == ['name', 'char']


def test_fieldsclasstype():
    f = Field(String())

    class Character(FieldsClassType):
        field_name = f

    assert Character._meta.fields == [f]


def test_fieldsclasstype_fieldtype():
    f = Field(String())

    class Character(FieldsClassType):
        field_name = f

    schema = Schema(query=Character)
    assert Character.fields_internal_types(schema)['fieldName'] == schema.T(f)
    assert Character._meta.fields_map['field_name'] == f


def test_fieldsclasstype_inheritfields():
    name_field = Field(String())
    last_name_field = Field(String())

    class Fields1(FieldsClassType):
        name = name_field

    class Fields2(Fields1):
        last_name = last_name_field

    assert list(Fields2._meta.fields_map.keys()) == ['name', 'last_name']




from graphql.core.type import GraphQLObjectType

from graphene.core.schema import Schema
from graphene.core.types import String

from ...types.argument import ArgumentsGroup
from ..mutation import Mutation


def test_mutation():
    class MyMutation(Mutation):
        '''MyMutation description'''
        class Input:
            arg_name = String()
        name = String()

    schema = Schema()

    object_type = schema.T(MyMutation)
    assert MyMutation._meta.type_name == 'MyMutation'
    assert isinstance(object_type, GraphQLObjectType)
    assert object_type.description == 'MyMutation description'
    assert list(object_type.get_fields().keys()) == ['name']
    assert MyMutation._meta.fields_map['name'].object_type == MyMutation
    assert isinstance(MyMutation.arguments, ArgumentsGroup)
    assert 'argName' in MyMutation.arguments



from .fields import (
    ConnectionField,
    NodeField,
    GlobalIDField,
)

from .types import (
    Node,
    PageInfo,
    Edge,
    Connection,
    ClientIDMutation
)

from .utils import is_node

__all__ = ['ConnectionField', 'NodeField', 'GlobalIDField', 'Node',
           'PageInfo', 'Edge', 'Connection', 'ClientIDMutation', 'is_node']



import six

from .objecttype import ObjectType, ObjectTypeMeta


class MutationMeta(ObjectTypeMeta):

    def construct(cls, bases, attrs):
        input_class = attrs.pop('Input', None)
        if input_class:
            items = dict(vars(input_class))
            items.pop('__dict__', None)
            items.pop('__doc__', None)
            items.pop('__module__', None)
            items.pop('__weakref__', None)
            cls.add_to_class('arguments', cls.construct_arguments(items))
        cls = super(MutationMeta, cls).construct(bases, attrs)
        return cls

    def construct_arguments(cls, items):
        from ..types.argument import ArgumentsGroup
        return ArgumentsGroup(**items)


class Mutation(six.with_metaclass(MutationMeta, ObjectType)):

    class Meta:
        abstract = True

    @classmethod
    def get_arguments(cls):
        return cls.arguments



from py.test import raises

from ..lazylist import LazyList


def test_lazymap():
    data = list(range(10))
    lm = LazyList(data)
    assert len(lm) == 10
    assert lm[1] == 1
    assert isinstance(lm[1:4], LazyList)
    assert lm.append == data.append
    assert repr(lm) == '<LazyList [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]>'


def test_lazymap_iter():
    data = list(range(2))
    lm = LazyList(data)
    iter_lm = iter(lm)
    assert iter_lm.next() == 0
    assert iter_lm.next() == 1
    with raises(StopIteration):
        iter_lm.next()



from graphql.core.type import GraphQLInputObjectField

import graphene
from graphene import relay
from graphene.core.schema import Schema

my_id = 0


class Query(graphene.ObjectType):
    base = graphene.String()


class ChangeNumber(relay.ClientIDMutation):
    '''Result mutation'''
    class Input:
        to = graphene.Int()

    result = graphene.String()

    @classmethod
    def mutate_and_get_payload(cls, input, info):
        global my_id
        my_id = input.get('to', my_id + 1)
        return ChangeNumber(result=my_id)


class MyResultMutation(graphene.ObjectType):
    change_number = graphene.Field(ChangeNumber)


schema = Schema(query=Query, mutation=MyResultMutation)


def test_mutation_arguments():
    assert ChangeNumber.arguments
    assert list(ChangeNumber.arguments) == ['input']
    assert 'input' in ChangeNumber.arguments
    inner_type = ChangeNumber.input_type
    client_mutation_id_field = inner_type._meta.fields_map[
        'client_mutation_id']
    assert issubclass(inner_type, graphene.InputObjectType)
    assert isinstance(client_mutation_id_field.type, graphene.NonNull)
    assert isinstance(client_mutation_id_field.type.of_type, graphene.String)
    assert client_mutation_id_field.object_type == inner_type
    assert isinstance(schema.T(client_mutation_id_field), GraphQLInputObjectField)


def test_execute_mutations():
    query = '''
    mutation M{
      first: changeNumber(input: {clientMutationId: "mutation1"}) {
        clientMutationId
        result
      },
      second: changeNumber(input: {clientMutationId: "mutation2"}) {
        clientMutationId
        result
      }
      third: changeNumber(input: {clientMutationId: "mutation3", to: 5}) {
        result
        clientMutationId
      }
    }
    '''
    expected = {
        'first': {
            'clientMutationId': 'mutation1',
            'result': '1',
        },
        'second': {
            'clientMutationId': 'mutation2',
            'result': '2',
        },
        'third': {
            'clientMutationId': 'mutation3',
            'result': '5',
        }
    }
    result = schema.execute(query, root=object())
    assert not result.errors
    assert result.data == expected



from django.db import models
from django.db.models.manager import Manager
from django.db.models.query import QuerySet

from graphene.utils import LazyList


def get_type_for_model(schema, model):
    schema = schema
    types = schema.types.values()
    for _type in types:
        type_model = hasattr(_type, '_meta') and getattr(
            _type._meta, 'model', None)
        if model == type_model:
            return _type


def get_reverse_fields(model):
    for name, attr in model.__dict__.items():
        related = getattr(attr, 'related', None)
        if isinstance(related, models.ManyToOneRel):
            yield related


class WrappedQueryset(LazyList):

    def __len__(self):
        # Dont calculate the length using len(queryset), as this will
        # evaluate the whole queryset and return it's length.
        # Use .count() instead
        return self._origin.count()


def maybe_queryset(value):
    if isinstance(value, Manager):
        value = value.get_queryset()
    if isinstance(value, QuerySet):
        return WrappedQueryset(value)
    return value



import six
from graphql.core.type import GraphQLList, GraphQLNonNull

from .base import LazyType, MountedType, MountType


class OfType(MountedType):

    def __init__(self, of_type, *args, **kwargs):
        if isinstance(of_type, six.string_types):
            of_type = LazyType(of_type)
        self.of_type = of_type
        super(OfType, self).__init__(*args, **kwargs)

    def internal_type(self, schema):
        return self.T(schema.T(self.of_type))

    def mount(self, cls):
        self.parent = cls
        if isinstance(self.of_type, MountType):
            self.of_type.mount(cls)


class List(OfType):
    T = GraphQLList


class NonNull(OfType):
    T = GraphQLNonNull



from graphql_relay.node.node import from_global_id

from ..core.fields import Field
from ..core.types.definitions import NonNull
from ..core.types.scalars import ID, Int, String


class ConnectionField(Field):

    def __init__(self, type, resolver=None, description='',
                 connection_type=None, edge_type=None, **kwargs):
        super(
            ConnectionField,
            self).__init__(
            type,
            resolver=resolver,
            before=String(),
            after=String(),
            first=Int(),
            last=Int(),
            description=description,
            **kwargs)
        self.connection_type = connection_type
        self.edge_type = edge_type

    def wrap_resolved(self, value, instance, args, info):
        return value

    def resolver(self, instance, args, info):
        schema = info.schema.graphene_schema
        connection_type = self.get_type(schema)
        resolved = super(ConnectionField, self).resolver(instance, args, info)
        if isinstance(resolved, connection_type):
            return resolved
        return connection_type.from_list(resolved, args, info)

    def get_connection_type(self, node):
        connection_type = self.connection_type or node.get_connection_type()
        edge_type = self.get_edge_type(node)
        return connection_type.for_node(node, edge_type=edge_type)

    def get_edge_type(self, node):
        edge_type = self.edge_type or node.get_edge_type()
        return edge_type.for_node(node)

    def get_type(self, schema):
        from graphene.relay.utils import is_node
        type = schema.T(self.type)
        node = schema.objecttype(type)
        assert is_node(node), 'Only nodes have connections.'
        schema.register(node)
        connection_type = self.get_connection_type(node)

        return connection_type


class NodeField(Field):
    '''Fetches an object given its ID'''

    def __init__(self, object_type=None, *args, **kwargs):
        from graphene.relay.types import Node
        id = kwargs.pop('id', None) or ID(description='The ID of an object')
        super(NodeField, self).__init__(
            object_type or Node, id=id, *args, **kwargs)
        self.field_object_type = object_type

    def id_fetcher(self, global_id, info):
        from graphene.relay.utils import is_node
        schema = info.schema.graphene_schema
        try:
            resolved_global_id = from_global_id(global_id)
        except:
            return None
        _type, _id = resolved_global_id.type, resolved_global_id.id
        object_type = schema.get_type(_type)
        if not is_node(object_type) or (self.field_object_type and
                                        object_type != self.field_object_type):
            return

        return object_type.get_node(_id, info)

    def resolver(self, instance, args, info):
        global_id = args.get('id')
        return self.id_fetcher(global_id, info)


class GlobalIDField(Field):
    '''The ID of an object'''

    def __init__(self, *args, **kwargs):
        super(GlobalIDField, self).__init__(NonNull(ID()), *args, **kwargs)

    def resolver(self, instance, args, info):
        return instance.to_global_id()



from graphql.core.type import GraphQLArgument
from pytest import raises

from graphene.core.schema import Schema
from graphene.core.types import ObjectType

from ..argument import Argument, snake_case_args, to_arguments
from ..scalars import String


def test_argument_internal_type():
    class MyObjectType(ObjectType):
        pass
    schema = Schema(query=MyObjectType)
    a = Argument(MyObjectType, description='My argument', default='3')
    type = schema.T(a)
    assert isinstance(type, GraphQLArgument)
    assert type.description == 'My argument'
    assert type.default_value == '3'


def test_to_arguments():
    arguments = to_arguments(
        Argument(String, name='myArg'),
        String(name='otherArg'),
        my_kwarg=String(),
        other_kwarg=String(),
    )

    assert [a.name for a in arguments] == [
        'myArg', 'otherArg', 'myKwarg', 'otherKwarg']


def test_to_arguments_no_name():
    with raises(AssertionError) as excinfo:
        to_arguments(
            String(),
        )
    assert 'must have a name' in str(excinfo.value)


def test_to_arguments_wrong_type():
    with raises(ValueError) as excinfo:
        to_arguments(
            p=3
        )
    assert 'Unknown argument p=3' == str(excinfo.value)


def test_snake_case_args():
    resolver = lambda instance, args, info: args['my_arg']['inner_arg']
    r = snake_case_args(resolver)
    assert r(None, {'myArg': {'innerArg': 3}}, None) == 3



from .types import Node


def is_node(object_type):
    return object_type and issubclass(
        object_type, Node) and not object_type._meta.abstract


def is_node_type(object_type):
    return object_type and issubclass(
        object_type, Node) and object_type._meta.abstract



from .models import Character, Faction, Ship


def initialize():
    human = Character(
        name='Human'
    )
    human.save()

    droid = Character(
        name='Droid'
    )
    droid.save()

    rebels = Faction(
        id='1',
        name='Alliance to Restore the Republic',
        hero=human
    )
    rebels.save()

    empire = Faction(
        id='2',
        name='Galactic Empire',
        hero=droid
    )
    empire.save()

    xwing = Ship(
        id='1',
        name='X-Wing',
        faction=rebels,
    )
    xwing.save()

    ywing = Ship(
        id='2',
        name='Y-Wing',
        faction=rebels,
    )
    ywing.save()

    awing = Ship(
        id='3',
        name='A-Wing',
        faction=rebels,
    )
    awing.save()

    # Yeah, technically it's Corellian. But it flew in the service of the rebels,
    # so for the purposes of this demo it's a rebel ship.
    falcon = Ship(
        id='4',
        name='Millenium Falcon',
        faction=rebels,
    )
    falcon.save()

    homeOne = Ship(
        id='5',
        name='Home One',
        faction=rebels,
    )
    homeOne.save()

    tieFighter = Ship(
        id='6',
        name='TIE Fighter',
        faction=empire,
    )
    tieFighter.save()

    tieInterceptor = Ship(
        id='7',
        name='TIE Interceptor',
        faction=empire,
    )
    tieInterceptor.save()

    executor = Ship(
        id='8',
        name='Executor',
        faction=empire,
    )
    executor.save()


def create_ship(ship_name, faction_id):
    new_ship = Ship(
        name=ship_name,
        faction_id=faction_id
    )
    new_ship.save()
    return new_ship


def get_ship(_id):
    return Ship.objects.get(id=_id)


def get_ships():
    return Ship.objects.all()


def get_faction(_id):
    return Faction.objects.get(id=_id)


def get_rebels():
    return get_faction(1)


def get_empire():
    return get_faction(2)



from django.conf.urls import url

import graphene
from graphene import Schema
from graphene.contrib.django.types import DjangoNode
from graphene.contrib.django.views import GraphQLView

from .models import Article, Reporter


class Character(DjangoNode):

    class Meta:
        model = Reporter

    def get_node(self, id):
        pass


class Human(DjangoNode):
    raises = graphene.String()

    class Meta:
        model = Article

    def resolve_raises(self, *args):
        raise Exception("This field should raise exception")

    def get_node(self, id):
        pass

schema = Schema(query=Human)


urlpatterns = [
    url(r'^graphql', GraphQLView.as_view(schema=schema)),
]



import collections

from .str_converters import to_camel_case, to_snake_case


class ProxySnakeDict(collections.MutableMapping):
    __slots__ = ('data')

    def __init__(self, data):
        self.data = data

    def __contains__(self, key):
        return key in self.data or to_camel_case(key) in self.data

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def __iter__(self):
        return self.iterkeys()

    def __len__(self):
        return len(self.data)

    def __delitem__(self, item):
        raise TypeError('ProxySnakeDict does not support item deletion')

    def __setitem__(self, item, value):
        raise TypeError('ProxySnakeDict does not support item assignment')

    def __getitem__(self, key):
        if key in self.data:
            item = self.data[key]
        else:
            camel_key = to_camel_case(key)
            if camel_key in self.data:
                item = self.data[camel_key]
            else:
                raise KeyError(key, camel_key)

        if isinstance(item, dict):
            return ProxySnakeDict(item)
        return item

    def keys(self):
        return list(self.iterkeys())

    def items(self):
        return list(self.iteritems())

    def iterkeys(self):
        for k in self.data.keys():
            yield to_snake_case(k)
        return

    def iteritems(self):
        for k in self.iterkeys():
            yield k, self[k]

    def to_data_dict(self):
        return self.data.__class__(self.iteritems())

    def __eq__(self, other):
        return self.to_data_dict() == other.to_data_dict()

    def __repr__(self):
        data_repr = self.to_data_dict().__repr__()
        return '<ProxySnakeDict {}>'.format(data_repr)



from py.test import raises

from graphene.core.classtypes import Options


class Meta:
    type_name = 'Character'


class InvalidMeta:
    other_value = True


def test_options_contribute():
    opt = Options(Meta)

    class ObjectType(object):
        pass

    opt.contribute_to_class(ObjectType, '_meta')
    assert ObjectType._meta == opt


def test_options_typename():
    opt = Options(Meta)

    class ObjectType(object):
        pass

    opt.contribute_to_class(ObjectType, '_meta')
    assert opt.type_name == 'Character'


def test_options_description():
    opt = Options(Meta)

    class ObjectType(object):

        '''False description'''

    opt.contribute_to_class(ObjectType, '_meta')
    assert opt.description == 'False description'


def test_field_no_contributed_raises_error():
    opt = Options(InvalidMeta)

    class ObjectType(object):
        pass

    with raises(Exception) as excinfo:
        opt.contribute_to_class(ObjectType, '_meta')

    assert 'invalid attribute' in str(excinfo.value)



try:
    from blinker import Signal
except ImportError:
    class Signal(object):

        def send(self, *args, **kwargs):
            pass

init_schema = Signal()
class_prepared = Signal()
pre_init = Signal()
post_init = Signal()



from graphql.core.type import (
    GraphQLEnumType as Enum
)

from graphene import signals

from .core import (
    Schema,
    ObjectType,
    InputObjectType,
    Interface,
    Mutation,
    Scalar,
    BaseType,
    LazyType,
    Argument,
    Field,
    InputField,
    String,
    Int,
    Boolean,
    ID,
    Float,
    List,
    NonNull
)

from graphene.core.fields import (
    StringField,
    IntField,
    BooleanField,
    IDField,
    ListField,
    NonNullField,
    FloatField,
)

from graphene.utils import (
    resolve_only_args
)

__all__ = [
    'Enum',
    'Argument',
    'String',
    'Int',
    'Boolean',
    'Float',
    'ID',
    'List',
    'NonNull',
    'signals',
    'Schema',
    'BaseType',
    'LazyType',
    'ObjectType',
    'InputObjectType',
    'Interface',
    'Mutation',
    'Scalar',
    'Field',
    'InputField',
    'StringField',
    'IntField',
    'BooleanField',
    'IDField',
    'ListField',
    'NonNullField',
    'FloatField',
    'resolve_only_args']



class LazyList(object):

    def __init__(self, origin, state=None):
        self._origin = origin
        self._state = state or []
        self._origin_iter = None
        self._finished = False

    def __iter__(self):
        return self if not self._finished else iter(self._state)

    def iter(self):
        return self.__iter__()

    def __len__(self):
        return self._origin.__len__()

    def __next__(self):
        try:
            if not self._origin_iter:
                self._origin_iter = self._origin.__iter__()
            n = next(self._origin_iter)
        except StopIteration as e:
            self._finished = True
            raise e
        else:
            self._state.append(n)
            return n

    def next(self):
        return self.__next__()

    def __getitem__(self, key):
        item = self._origin[key]
        if isinstance(key, slice):
            return self.__class__(item)
        return item

    def __getattr__(self, name):
        return getattr(self._origin, name)

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, repr(self._origin))



from graphql.core.type import GraphQLEnumValue

import graphene
from graphene import resolve_only_args

from .data import get_character, get_droid, get_hero, get_human

Episode = graphene.Enum('Episode', dict(
    NEWHOPE=GraphQLEnumValue(4),
    EMPIRE=GraphQLEnumValue(5),
    JEDI=GraphQLEnumValue(6)
))


class Character(graphene.Interface):
    id = graphene.ID()
    name = graphene.String()
    friends = graphene.List('Character')
    appears_in = graphene.List(Episode)

    def resolve_friends(self, args, *_):
        # The character friends is a list of strings
        return [get_character(f) for f in self.friends]


class Human(Character):
    home_planet = graphene.String()


class Droid(Character):
    primary_function = graphene.String()


class Query(graphene.ObjectType):
    hero = graphene.Field(Character,
                          episode=graphene.Argument(Episode)
                          )
    human = graphene.Field(Human,
                           id=graphene.String()
                           )
    droid = graphene.Field(Droid,
                           id=graphene.String()
                           )

    @resolve_only_args
    def resolve_hero(self, episode=None):
        return get_hero(episode)

    @resolve_only_args
    def resolve_human(self, id):
        return get_human(id)

    @resolve_only_args
    def resolve_droid(self, id):
        return get_droid(id)


Schema = graphene.Schema(query=Query)



from graphql.core.type import GraphQLList
from pytest import raises

import graphene
from graphene import relay

schema = graphene.Schema()


class OtherNode(relay.Node):
    name = graphene.String()

    @classmethod
    def get_node(cls, id, info):
        pass


def test_works_old_get_node():
    class Part(relay.Node):
        x = graphene.String()

        @classmethod
        def get_node(cls, id):
            return id

    assert Part.get_node(1) == 1


def test_works_old_static_get_node():
    class Part(relay.Node):
        x = graphene.String()

        @staticmethod
        def get_node(id):
            return id

    assert Part.get_node(1) == 1


def test_field_no_contributed_raises_error():
    with raises(Exception) as excinfo:
        class Part(relay.Node):
            x = graphene.String()

    assert 'get_node' in str(excinfo.value)


def test_node_should_have_same_connection_always():
    connection1 = relay.Connection.for_node(OtherNode)
    connection2 = relay.Connection.for_node(OtherNode)

    assert connection1 == connection2


def test_node_should_have_id_field():
    assert 'id' in OtherNode._meta.fields_map


def test_node_connection_should_have_edge():
    connection = relay.Connection.for_node(OtherNode)
    edge = relay.Edge.for_node(OtherNode)
    connection_type = schema.T(connection)
    connection_fields = connection_type.get_fields()
    assert 'edges' in connection_fields
    assert 'pageInfo' in connection_fields
    edges_type = connection_fields['edges'].type
    assert isinstance(edges_type, GraphQLList)
    assert edges_type.of_type == schema.T(edge)



from graphql.core.type import GraphQLUnionType

from graphene.core.schema import Schema
from graphene.core.types import String

from ..objecttype import ObjectType
from ..uniontype import UnionType


def test_uniontype():
    class Human(ObjectType):
        name = String()

    class Pet(ObjectType):
        name = String()

    class Thing(UnionType):
        '''Thing union description'''
        class Meta:
            types = [Human, Pet]

    schema = Schema()

    object_type = schema.T(Thing)
    assert isinstance(object_type, GraphQLUnionType)
    assert Thing._meta.type_name == 'Thing'
    assert object_type.description == 'Thing union description'
    assert object_type.get_possible_types() == [schema.T(Human), schema.T(Pet)]



from ..str_converters import to_camel_case, to_snake_case


def test_snake_case():
    assert to_snake_case('snakesOnAPlane') == 'snakes_on_a_plane'
    assert to_snake_case('SnakesOnAPlane') == 'snakes_on_a_plane'
    assert to_snake_case('snakes_on_a_plane') == 'snakes_on_a_plane'
    assert to_snake_case('IPhoneHysteria') == 'i_phone_hysteria'
    assert to_snake_case('iPhoneHysteria') == 'i_phone_hysteria'


def test_camel_case():
    assert to_camel_case('snakes_on_a_plane') == 'snakesOnAPlane'
    assert to_camel_case('i_phone_hysteria') == 'iPhoneHysteria'



from functools import partial

import six
from graphql.core.type import GraphQLObjectType

from graphene import signals

from .base import FieldsClassType, FieldsClassTypeMeta, FieldsOptions
from .uniontype import UnionType


def is_objecttype(cls):
    if not issubclass(cls, ObjectType):
        return False
    return not cls._meta.interface


class ObjectTypeOptions(FieldsOptions):

    def __init__(self, *args, **kwargs):
        super(ObjectTypeOptions, self).__init__(*args, **kwargs)
        self.interface = False
        self.interfaces = []


class ObjectTypeMeta(FieldsClassTypeMeta):

    def construct(cls, bases, attrs):
        cls = super(ObjectTypeMeta, cls).construct(bases, attrs)
        if not cls._meta.abstract:
            union_types = list(filter(is_objecttype, bases))
            if len(union_types) > 1:
                meta_attrs = dict(cls._meta.original_attrs, types=union_types)
                Meta = type('Meta', (object, ), meta_attrs)
                attrs['Meta'] = Meta
                attrs['__module__'] = cls.__module__
                attrs['__doc__'] = cls.__doc__
                return type(cls.__name__, (UnionType, ), attrs)
        return cls

    options_class = ObjectTypeOptions


class ObjectType(six.with_metaclass(ObjectTypeMeta, FieldsClassType)):

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        signals.pre_init.send(self.__class__, args=args, kwargs=kwargs)
        self._root = kwargs.pop('_root', None)
        args_len = len(args)
        fields = self._meta.fields
        if args_len > len(fields):
            # Daft, but matches old exception sans the err msg.
            raise IndexError("Number of args exceeds number of fields")
        fields_iter = iter(fields)

        if not kwargs:
            for val, field in zip(args, fields_iter):
                setattr(self, field.attname, val)
        else:
            for val, field in zip(args, fields_iter):
                setattr(self, field.attname, val)
                kwargs.pop(field.attname, None)

        for field in fields_iter:
            try:
                val = kwargs.pop(field.attname)
                setattr(self, field.attname, val)
            except KeyError:
                pass

        if kwargs:
            for prop in list(kwargs):
                try:
                    if isinstance(getattr(self.__class__, prop), property):
                        setattr(self, prop, kwargs.pop(prop))
                except AttributeError:
                    pass
            if kwargs:
                raise TypeError(
                    "'%s' is an invalid keyword argument for this function" %
                    list(kwargs)[0])

        signals.post_init.send(self.__class__, instance=self)

    @classmethod
    def internal_type(cls, schema):
        if cls._meta.abstract:
            raise Exception("Abstract ObjectTypes don't have a specific type.")

        return GraphQLObjectType(
            cls._meta.type_name,
            description=cls._meta.description,
            interfaces=list(map(schema.T, cls._meta.interfaces)),
            fields=partial(cls.fields_internal_types, schema),
            is_type_of=getattr(cls, 'is_type_of', None)
        )

    @classmethod
    def wrap(cls, instance, args, info):
        return cls(_root=instance)



from graphql.core.type import GraphQLScalarType

from ...schema import Schema
from ..scalar import Scalar


def test_custom_scalar():
    import datetime
    from graphql.core.language import ast

    class DateTimeScalar(Scalar):
        '''DateTimeScalar Documentation'''
        @staticmethod
        def serialize(dt):
            return dt.isoformat()

        @staticmethod
        def parse_literal(node):
            if isinstance(node, ast.StringValue):
                return datetime.datetime.strptime(
                    node.value, "%Y-%m-%dT%H:%M:%S.%f")

        @staticmethod
        def parse_value(value):
            return datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f")

    schema = Schema()

    scalar_type = schema.T(DateTimeScalar)
    assert isinstance(scalar_type, GraphQLScalarType)
    assert scalar_type.name == 'DateTimeScalar'
    assert scalar_type.description == 'DateTimeScalar Documentation'



from .schema import (
    Schema
)

from .classtypes import (
    ObjectType,
    InputObjectType,
    Interface,
    Mutation,
    Scalar
)

from .types import (
    BaseType,
    LazyType,
    Argument,
    Field,
    InputField,
    String,
    Int,
    Boolean,
    ID,
    Float,
    List,
    NonNull
)

__all__ = [
    'Argument',
    'String',
    'Int',
    'Boolean',
    'Float',
    'ID',
    'List',
    'NonNull',
    'Schema',
    'BaseType',
    'LazyType',
    'ObjectType',
    'InputObjectType',
    'Interface',
    'Mutation',
    'Scalar',
    'Field',
    'InputField']



from functools import partial

import six
from graphql.core.type import GraphQLInterfaceType

from .base import FieldsClassTypeMeta
from .objecttype import ObjectType, ObjectTypeMeta


class InterfaceMeta(ObjectTypeMeta):

    def construct(cls, bases, attrs):
        if cls._meta.abstract or Interface in bases:
            # Return Interface type
            cls = FieldsClassTypeMeta.construct(cls, bases, attrs)
            setattr(cls._meta, 'interface', True)
            return cls
        else:
            # Return ObjectType class with all the inherited interfaces
            cls = super(InterfaceMeta, cls).construct(bases, attrs)
            for interface in bases:
                is_interface = issubclass(interface, Interface) and getattr(interface._meta, 'interface', False)
                if not is_interface:
                    continue
                cls._meta.interfaces.append(interface)
            return cls


class Interface(six.with_metaclass(InterfaceMeta, ObjectType)):

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        if self._meta.interface:
            raise Exception("An interface cannot be initialized")
        return super(Interface, self).__init__(*args, **kwargs)

    @classmethod
    def _resolve_type(cls, schema, instance, *args):
        return schema.T(instance.__class__)

    @classmethod
    def internal_type(cls, schema):
        if not cls._meta.interface:
            return super(Interface, cls).internal_type(schema)

        return GraphQLInterfaceType(
            cls._meta.type_name,
            description=cls._meta.description,
            resolve_type=partial(cls._resolve_type, schema),
            fields=partial(cls.fields_internal_types, schema)
        )



from collections import OrderedDict
from functools import wraps

import six
from graphql.core.type import GraphQLField, GraphQLInputObjectField

from ...utils import to_camel_case
from ..classtypes.base import FieldsClassType
from ..classtypes.inputobjecttype import InputObjectType
from ..classtypes.mutation import Mutation
from .argument import ArgumentsGroup, snake_case_args
from .base import LazyType, MountType, OrderedType
from .definitions import NonNull


class Field(OrderedType):

    def __init__(
            self, type, description=None, args=None, name=None, resolver=None,
            required=False, default=None, *args_list, **kwargs):
        _creation_counter = kwargs.pop('_creation_counter', None)
        super(Field, self).__init__(_creation_counter=_creation_counter)
        self.name = name
        if isinstance(type, six.string_types):
            type = LazyType(type)
        self.required = required
        self.type = type
        self.description = description
        args = OrderedDict(args or {}, **kwargs)
        self.arguments = ArgumentsGroup(*args_list, **args)
        self.object_type = None
        self.resolver_fn = resolver
        self.default = default

    def contribute_to_class(self, cls, attname):
        assert issubclass(
            cls, (FieldsClassType)), 'Field {} cannot be mounted in {}'.format(
            self, cls)
        if not self.name:
            self.name = to_camel_case(attname)
        self.attname = attname
        self.object_type = cls
        self.mount(cls)
        if isinstance(self.type, MountType):
            self.type.mount(cls)
        cls._meta.add_field(self)

    @property
    def resolver(self):
        return self.resolver_fn or self.get_resolver_fn()

    def get_resolver_fn(self):
        resolve_fn_name = 'resolve_%s' % self.attname
        if hasattr(self.object_type, resolve_fn_name):
            return getattr(self.object_type, resolve_fn_name)

        def default_getter(instance, args, info):
            return getattr(instance, self.attname, self.default)
        return default_getter

    def get_type(self, schema):
        if self.required:
            return NonNull(self.type)
        return self.type

    def decorate_resolver(self, resolver):
        return snake_case_args(resolver)

    def internal_type(self, schema):
        resolver = self.resolver
        description = self.description
        arguments = self.arguments
        if not description and resolver:
            description = resolver.__doc__
        type = schema.T(self.get_type(schema))
        type_objecttype = schema.objecttype(type)
        if type_objecttype and issubclass(type_objecttype, Mutation):
            assert len(arguments) == 0
            arguments = type_objecttype.get_arguments()
            resolver = getattr(type_objecttype, 'mutate')
        else:
            my_resolver = resolver

            @wraps(my_resolver)
            def wrapped_func(instance, args, info):
                if not isinstance(instance, self.object_type):
                    instance = self.object_type(_root=instance)
                return my_resolver(instance, args, info)
            resolver = wrapped_func

        assert type, 'Internal type for field %s is None' % str(self)
        return GraphQLField(type, args=schema.T(arguments),
                            resolver=self.decorate_resolver(resolver),
                            description=description,)

    def __repr__(self):
        """
        Displays the module, class and name of the field.
        """
        path = '%s.%s' % (self.__class__.__module__, self.__class__.__name__)
        name = getattr(self, 'attname', None)
        if name is not None:
            return '<%s: %s>' % (path, name)
        return '<%s>' % path

    def __str__(self):
        """ Return "object_type.field_name". """
        return '%s.%s' % (self.object_type.__name__, self.attname)

    def __eq__(self, other):
        eq = super(Field, self).__eq__(other)
        if isinstance(self, type(other)):
            return eq and self.object_type == other.object_type
        return NotImplemented

    def __hash__(self):
        return hash((self.creation_counter, self.object_type))


class InputField(OrderedType):

    def __init__(self, type, description=None, default=None,
                 name=None, _creation_counter=None, required=False):
        super(InputField, self).__init__(_creation_counter=_creation_counter)
        self.name = name
        if required:
            type = NonNull(type)
        self.type = type
        self.description = description
        self.default = default

    def contribute_to_class(self, cls, attname):
        assert issubclass(
            cls, (InputObjectType)), 'InputField {} cannot be mounted in {}'.format(
            self, cls)
        if not self.name:
            self.name = to_camel_case(attname)
        self.attname = attname
        self.object_type = cls
        self.mount(cls)
        if isinstance(self.type, MountType):
            self.type.mount(cls)
        cls._meta.add_field(self)

    def internal_type(self, schema):
        return GraphQLInputObjectField(
            schema.T(self.type),
            default_value=self.default, description=self.description)



import graphene


class Patron(graphene.ObjectType):
    id = graphene.ID()
    name = graphene.String()
    age = graphene.ID()


class Query(graphene.ObjectType):

    patron = graphene.Field(Patron)

    def resolve_patron(self, args, info):
        return Patron(id=1, name='Demo')

schema = graphene.Schema(query=Query)
query = '''
    query something{
      patron {
        id
        name
      }
    }
'''
result = schema.execute(query)
print(result.data['patron'])



import sys

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

setup(
    name='graphene',
    version='0.4.3',

    description='Graphene: Python DSL for GraphQL',
    long_description=open('README.rst').read(),

    url='https://github.com/graphql-python/graphene',

    author='Syrus Akbary',
    author_email='me@syrusakbary.com',

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],

    keywords='api graphql protocol rest relay graphene',

    packages=find_packages(exclude=['tests']),

    install_requires=[
        'six>=1.10.0',
        'graphql-core==0.4.9',
        'graphql-relay==0.3.3'
    ],
    tests_require=[
        'pytest>=2.7.2',
        'pytest-django',
        'mock',
    ],
    extras_require={
        'django': [
            'Django>=1.6.0,<1.9',
            'singledispatch>=3.4.0.3',
            'graphql-django-view>=1.0.0',
        ],
    },

    cmdclass={'test': PyTest},
)



from py.test import raises

from graphene.contrib.django import DjangoObjectType
from tests.utils import assert_equal_lists

from .models import Reporter


def test_should_raise_if_no_model():
    with raises(Exception) as excinfo:
        class Character1(DjangoObjectType):
            pass
    assert 'model in the Meta' in str(excinfo.value)


def test_should_raise_if_model_is_invalid():
    with raises(Exception) as excinfo:
        class Character2(DjangoObjectType):

            class Meta:
                model = 1
    assert 'not a Django model' in str(excinfo.value)


def test_should_map_fields_correctly():
    class ReporterType2(DjangoObjectType):

        class Meta:
            model = Reporter
    assert_equal_lists(
        ReporterType2._meta.fields_map.keys(),
        ['articles', 'first_name', 'last_name', 'email', 'pets', 'id']
    )


def test_should_map_only_few_fields():
    class Reporter2(DjangoObjectType):

        class Meta:
            model = Reporter
            only_fields = ('id', 'email')
    assert_equal_lists(
        Reporter2._meta.fields_map.keys(),
        ['id', 'email']
    )



from graphql.core.type import GraphQLObjectType
from mock import patch

from graphene import Schema
from graphene.contrib.django.types import DjangoNode, DjangoObjectType
from graphene.core.fields import Field
from graphene.core.types.scalars import Int
from graphene.relay.fields import GlobalIDField
from tests.utils import assert_equal_lists

from .models import Article, Reporter

schema = Schema()


@schema.register
class Character(DjangoObjectType):
    '''Character description'''
    class Meta:
        model = Reporter


@schema.register
class Human(DjangoNode):
    '''Human description'''

    pub_date = Int()

    class Meta:
        model = Article


def test_django_interface():
    assert DjangoNode._meta.interface is True


@patch('graphene.contrib.django.tests.models.Article.objects.get', return_value=Article(id=1))
def test_django_get_node(get):
    human = Human.get_node(1, None)
    get.assert_called_with(id=1)
    assert human.id == 1


def test_djangonode_idfield():
    idfield = DjangoNode._meta.fields_map['id']
    assert isinstance(idfield, GlobalIDField)


def test_node_idfield():
    idfield = Human._meta.fields_map['id']
    assert isinstance(idfield, GlobalIDField)


def test_node_replacedfield():
    idfield = Human._meta.fields_map['pub_date']
    assert isinstance(idfield, Field)
    assert schema.T(idfield).type == schema.T(Int())


def test_objecttype_init_none():
    h = Human()
    assert h._root is None


def test_objecttype_init_good():
    instance = Article()
    h = Human(instance)
    assert h._root == instance


def test_object_type():
    object_type = schema.T(Human)
    Human._meta.fields_map
    assert Human._meta.interface is False
    assert isinstance(object_type, GraphQLObjectType)
    assert_equal_lists(
        object_type.get_fields().keys(),
        ['headline', 'id', 'reporter', 'pubDate']
    )
    assert schema.T(DjangoNode) in object_type.get_interfaces()


def test_node_notinterface():
    assert Human._meta.interface is False
    assert DjangoNode in Human._meta.interfaces



from .base import BaseType, LazyType, OrderedType
from .argument import Argument, ArgumentsGroup, to_arguments
from .definitions import List, NonNull
# Compatibility import
from .objecttype import Interface, ObjectType, Mutation, InputObjectType

from .scalars import String, ID, Boolean, Int, Float, Scalar
from .field import Field, InputField

__all__ = [
    'BaseType',
    'LazyType',
    'OrderedType',
    'Argument',
    'ArgumentsGroup',
    'to_arguments',
    'List',
    'NonNull',
    'Field',
    'InputField',
    'Interface',
    'ObjectType',
    'Mutation',
    'InputObjectType',
    'String',
    'ID',
    'Boolean',
    'Int',
    'Float',
    'Scalar']



import six
from graphql.core.type import GraphQLUnionType

from .base import FieldsClassType, FieldsClassTypeMeta, FieldsOptions


class UnionTypeOptions(FieldsOptions):

    def __init__(self, *args, **kwargs):
        super(UnionTypeOptions, self).__init__(*args, **kwargs)
        self.types = []


class UnionTypeMeta(FieldsClassTypeMeta):
    options_class = UnionTypeOptions

    def get_options(cls, meta):
        return cls.options_class(meta, types=[])


class UnionType(six.with_metaclass(UnionTypeMeta, FieldsClassType)):

    class Meta:
        abstract = True

    @classmethod
    def _resolve_type(cls, schema, instance, *args):
        return schema.T(instance.__class__)

    @classmethod
    def internal_type(cls, schema):
        if cls._meta.abstract:
            raise Exception("Abstract ObjectTypes don't have a specific type.")

        return GraphQLUnionType(
            cls._meta.type_name,
            types=list(map(schema.T, cls._meta.types)),
            resolve_type=cls._resolve_type,
            description=cls._meta.description,
        )



import inspect

import six
from django.db import models

from ...core.classtypes.objecttype import ObjectType, ObjectTypeMeta
from ...relay.types import Connection, Node, NodeMeta
from .converter import convert_django_field
from .options import DjangoOptions
from .utils import get_reverse_fields, maybe_queryset


class DjangoObjectTypeMeta(ObjectTypeMeta):
    options_class = DjangoOptions

    def construct_fields(cls):
        only_fields = cls._meta.only_fields
        reverse_fields = get_reverse_fields(cls._meta.model)
        all_fields = sorted(list(cls._meta.model._meta.fields) +
                            list(cls._meta.model._meta.local_many_to_many))
        all_fields += list(reverse_fields)
        already_created_fields = {f.attname for f in cls._meta.local_fields}

        for field in all_fields:
            is_not_in_only = only_fields and field.name not in only_fields
            is_already_created = field.name in already_created_fields
            is_excluded = field.name in cls._meta.exclude_fields or is_already_created
            if is_not_in_only or is_excluded:
                # We skip this field if we specify only_fields and is not
                # in there. Or when we exclude this field in exclude_fields
                continue
            converted_field = convert_django_field(field)
            cls.add_to_class(field.name, converted_field)

    def construct(cls, *args, **kwargs):
        cls = super(DjangoObjectTypeMeta, cls).construct(*args, **kwargs)
        if not cls._meta.abstract:
            if not cls._meta.model:
                raise Exception(
                    'Django ObjectType %s must have a model in the Meta class attr' %
                    cls)
            elif not inspect.isclass(cls._meta.model) or not issubclass(cls._meta.model, models.Model):
                raise Exception('Provided model in %s is not a Django model' % cls)

            cls.construct_fields()
        return cls


class InstanceObjectType(ObjectType):

    class Meta:
        abstract = True

    def __init__(self, _root=None):
        if _root:
            assert isinstance(_root, self._meta.model), (
                '{} received a non-compatible instance ({}) '
                'when expecting {}'.format(
                    self.__class__.__name__,
                    _root.__class__.__name__,
                    self._meta.model.__name__
                ))
        super(InstanceObjectType, self).__init__(_root=_root)

    @property
    def instance(self):
        return self._root

    @instance.setter
    def instance(self, value):
        self._root = value

    def __getattr__(self, attr):
        return getattr(self._root, attr)


class DjangoObjectType(six.with_metaclass(
        DjangoObjectTypeMeta, InstanceObjectType)):

    class Meta:
        abstract = True


class DjangoConnection(Connection):

    @classmethod
    def from_list(cls, iterable, *args, **kwargs):
        iterable = maybe_queryset(iterable)
        return super(DjangoConnection, cls).from_list(iterable, *args, **kwargs)


class DjangoNodeMeta(DjangoObjectTypeMeta, NodeMeta):
    pass


class NodeInstance(Node, InstanceObjectType):

    class Meta:
        abstract = True


class DjangoNode(six.with_metaclass(
        DjangoNodeMeta, NodeInstance)):

    class Meta:
        abstract = True

    @classmethod
    def get_node(cls, id, info=None):
        try:
            instance = cls._meta.model.objects.get(id=id)
            return cls(instance)
        except cls._meta.model.DoesNotExist:
            return None

    connection_type = DjangoConnection



from graphql.core.type import GraphQLObjectType
from py.test import raises

from graphene.core.schema import Schema
from graphene.core.types import String

from ..objecttype import ObjectType
from ..uniontype import UnionType


def test_object_type():
    class Human(ObjectType):
        '''Human description'''
        name = String()
        friends = String()

    schema = Schema()

    object_type = schema.T(Human)
    assert Human._meta.type_name == 'Human'
    assert isinstance(object_type, GraphQLObjectType)
    assert object_type.description == 'Human description'
    assert list(object_type.get_fields().keys()) == ['name', 'friends']
    assert Human._meta.fields_map['name'].object_type == Human


def test_object_type_container():
    class Human(ObjectType):
        name = String()
        friends = String()

    h = Human(name='My name')
    assert h.name == 'My name'


def test_object_type_set_properties():
    class Human(ObjectType):
        name = String()
        friends = String()

        @property
        def readonly_prop(self):
            return 'readonly'

        @property
        def write_prop(self):
            return self._write_prop

        @write_prop.setter
        def write_prop(self, value):
            self._write_prop = value

    h = Human(readonly_prop='custom', write_prop='custom')
    assert h.readonly_prop == 'readonly'
    assert h.write_prop == 'custom'


def test_object_type_container_invalid_kwarg():
    class Human(ObjectType):
        name = String()

    with raises(TypeError):
        Human(invalid='My name')


def test_object_type_container_too_many_args():
    class Human(ObjectType):
        name = String()

    with raises(IndexError):
        Human('Peter', 'No friends :(', None)


def test_object_type_union():
    class Human(ObjectType):
        name = String()

    class Pet(ObjectType):
        name = String()

    class Thing(Human, Pet):
        '''Thing union description'''
        my_attr = True

    assert issubclass(Thing, UnionType)
    assert Thing._meta.types == [Human, Pet]
    assert Thing._meta.type_name == 'Thing'
    assert Thing._meta.description == 'Thing union description'
    assert Thing.my_attr
